@@ -9,7 +9,7 @@ use Hash::FieldHash ':all';
 
 fieldhash my %graph => 'graph';
 
-our $VERSION = '2.09';
+our $VERSION = '2.10';
 
 # -----------------------------------------------
 
@@ -153,7 +153,7 @@ sub _init
 	$$arg{graph}    ||= GraphViz2 -> new
 		(
 		 edge   => {color => 'grey'},
-		 global => {directed => 1, record_orientation => 'horizontal'},
+		 global => {directed => 1},
 		 graph  => {rankdir => 'TB'},
 		 logger => '',
 		 node   => {color => 'blue', shape => 'oval'},
@@ -189,23 +189,23 @@ L<GraphViz2::Parse::Marpa> - Visualize a Marpa grammar as a graph
 =head1 Synopsis
 
 	#!/usr/bin/env perl
-	
+
 	use strict;
 	use warnings;
-	
+
 	use File::Spec;
-	
+
 	use GraphViz2;
 	use GraphViz2::Parse::Marpa;
-	
+
 	use Log::Handler;
-	
+
 	use Perl6::Slurp;
-	
+
 	# ------------------------------------------------
-	
+
 	my($logger) = Log::Handler -> new;
-	
+
 	$logger -> add
 		(
 		 screen =>
@@ -215,23 +215,23 @@ L<GraphViz2::Parse::Marpa> - Visualize a Marpa grammar as a graph
 			 minlevel       => 'error',
 		 }
 		);
-	
+
 	my($graph)  = GraphViz2 -> new
 		(
 		 edge   => {color => 'grey'},
-		 global => {directed => 1, record_orientation => 'horizontal'},
+		 global => {directed => 1},
 		 graph  => {rankdir => 'TB'},
 		 logger => $logger,
 		 node   => {color => 'blue', shape => 'oval'},
 		);
 	my($g)      = GraphViz2::Parse::Marpa -> new(graph => $graph);
 	my $grammar = slurp(File::Spec -> catfile('t', 'sample.marpa.1.dat') );
-	
+
 	$g -> create(grammar => $grammar);
-	
+
 	my($format)      = shift || 'svg';
 	my($output_file) = shift || File::Spec -> catfile('html', "parse.marpa.$format");
-	
+
 	$graph -> run(format => $format, output_file => $output_file);
 
 See scripts/parse.marpa.pl (L<GraphViz2/Scripts Shipped with this Module>).
